/* Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* System includes */
#include <stdio.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/timex.h>
#include <sys/timeb.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

/* Our includes */
#include "hash.h"
#include "constants.h"
#include "conn_stats.h"
#include "flow_stats.h"
#include "dward.config"
#include "dward.h"

/* Replay flag */
extern int REPLAY;

/* Statistical data */
long conns = 0;
long false_bad = 0;

/* Current time */
extern long timenow;

/* Connection ID for red-teaming */
int CONN_ID = 1;

/* Gather data on allocated memory, useful for checking if 
   everything is freed or there is memory leak */
int connmem = 0;

/* Debug file for conn statistics */
FILE* conndebug;

/* Debug flag */
extern int DEBUG;

/* Flags for statistics */
extern int STATS;

/* Statistics hash */
struct hash_table* ConnHash; 

int updates =  0, finds = 0, inserts = 0, gettings = 0;
 
/* Print connection protocol */
char* print_proto(int num)
{
  if (num == TCP)
    return "TCP";
  else if (num == UDP)
    return "UDP";
  else if (num == ICMP)
    return "ICMP";
  else
    return "OTHER";
}

/* Hash related functions we need to define */
int conn_clean(struct conn_stats* data)
{
  /* Hash entry should be cleaned if there is 
     no conn activity for a certain period. This period
     is different for good and transient conns. Good conns
     are kept longer in hash. */ 
  long curtime = time(0);
  if ((data->classification == GOOD &&
       curtime-data->lst > CONN_GOOD_INACTIVE_PERIOD && 
       curtime-data->lrt > CONN_GOOD_INACTIVE_PERIOD) ||
      (data->classification != GOOD &&
       curtime-data->lst > CONN_TRAN_INACTIVE_PERIOD && 
       curtime-data->lrt > CONN_TRAN_INACTIVE_PERIOD))
    return 1;
  else
    return 0;
}

int conn_compare(struct conn_key* k1, struct conn_key* k2)
{
  if (k1 == NULL || k2 == NULL)
    return FALSE;
  if (k1->localIP == k2->localIP &&
      k1->foreignIP == k2->foreignIP &&
      k1->localport == k2->localport &&
      k1->foreignport == k2->foreignport)
    return TRUE;
  else
    return FALSE;
}

unsigned long conn_hash1(struct conn_key* k, unsigned long size)
{
  if (k)
  /* Here we just hash sum of fields */
  return (k->localIP % size + k->foreignIP % size 
	  + k->localport + k->foreignport) % size;
  else
    return FALSE;
}

unsigned long conn_hash2(struct conn_key* k, unsigned long size)
{
  /* Here we just hash sum of fields */
  return (1+conn_hash1(k,size-1));
}
/* End of required hash related functions */

/* Get packet type in string format */
char* get_type(enum pack_type ptype)
{
  switch (ptype)
    {
    case SYN:
      return "SYN";
    case SYNACK:
      return "SYNACK";
    case ACK:
      return "ACK";
    case DP:
      return "DATA PACKET";
    case FIN:
      return "FIN";
    case FINACK:
      return "FINACK";
    case RST:
      return "RST";
    default: 
      return "UNKNOWN";
    }
}

/* Get connection state in string format */
char* get_state(enum app_state state)
{
  switch(state)
    {
    case NOSTATE:
      return "NONE";
    case ERROR:
      return "ERROR";
    case DNS_REQ:
      return "DNS_REQ";
    case DNS_RESP:
      return "DNS_RESP";
    case NTP_SENT:
      return "NTP_SENT";
    case NTP_REC:
      return "NTP_REC";
    case NTP_SENTREC:
      return "NTP_SENTREC";
    default:
      return "UNKNOWN";

    }
}

/* Both find functions are unsafe to use since they 
   return index to the entry that can be changed
   any moment due to rehashing or parallel access
   by other process */

/* Finds given connection in hash, 
   really just masks the fact that 
   we use hash for statistics bookkeeping 
   input: IP addresses and ports in char array 
   (addresses in dotted notation)			       
   output: index (-1 not found)	*/
long find_conn(char *address1, char* address2, char* port1, char* port2)
{
  struct in_addr in;  
  unsigned long a1, a2;
  int p1, p2;
  inet_aton(address1,&in);
  a1 = in.s_addr;
  inet_aton(address2,&in);
  a2 = in.s_addr;
  p1 = atoi(port1);
  p2 = atoi(port2);
  return find_conn_num(a1,a2,p1,p2); 
}

/* Finds connection, given partial data.
   Really puts an overhead on system since
   it traverses all connections to find the 
   needed one.
   input: unsigned int IP address, short known_port,
   unsigne int another IP address with unknown port
   output: pointer to data, 0 - not found */
struct conn_stats* find_partial_conn_num(unsigned int address1,   
					 short port1, unsigned int address2)
{
  long i;
  struct conn_key key;

  printf("Looking for %x and %x and %d\n",address1,address2,port1);
  for(i=0;i<ConnHash->size;i++)
    {
      /* For each valid connection */
      if(ConnHash->entries[i].state == VALID)
        {
	  key = *((struct conn_key*)ConnHash->entries[i].key);
	  printf("Checking %x:%d to %x:%d\n",key.localIP, key.localport,
		 key.foreignIP, key.foreignport);
	  if(key.localIP == address1 && key.foreignIP == address2
	     && key.foreignport == port1)
	      return (struct conn_stats*) ConnHash->entries[i].data;
	  if(key.localIP == address1 && key.foreignIP == address2
	     && key.localport == port1)
	      return (struct conn_stats*) ConnHash->entries[i].data;
	}
    }
  return 0;
}

/* Finds given connection in numeric form, 
   really just masks the fact that 
   we use hash for statistics bookkeeping 
   input: unsigned int IP addresses, short ports 			       
   output: index (-1 not found)	*/
long find_conn_num(unsigned int address1,  unsigned int address2, 
		   short port1, short port2)
{
  long index;
  struct conn_key key;
  finds ++;

  /* Create connection key */
  key.localIP = address1;
  key.foreignIP = address2;
  key.localport = port1;
  key.foreignport = port2;
  index = find(&key,ConnHash);
  return index;
}

/* Add connection to cache, return index 
   input: IP addresses and ports in char array 
   (addresses in dotted notation), protocol			       
   output: index (-1 not able to add) */ 
long add_conn(char* address1, char* address2, char* port1, char* port2,
	      enum conn_type proto)
{
  struct in_addr in;  
  unsigned long a1, a2;
  int p1, p2;
  inet_aton(address1,&in);
  a1 = in.s_addr;
  inet_aton(address2,&in);
  a2 = in.s_addr;
  p1 = atoi(port1);
  p2 = atoi(port2);
  return add_conn_num(a1,a2,p1,p2,proto);
}

/* Add connection to hash, return index 
   input: unsigned int IP addresses, short ports 			       
   output: index (-1 not found)	*/
long add_conn_num(unsigned int address1,  unsigned int address2, 
		   short port1, short port2, enum conn_type proto)
{
  long index,i;
  struct timeb* tm = (struct timeb*)malloc(sizeof(struct timeb));
  struct conn_stats* data;
  struct conn_key key;
  int outcome = 0;
  inserts++;

  connmem += sizeof(struct timeb);

  /* Create connection key */
  key.localIP = address1;
  key.foreignIP = address2;
  key.localport = port1;
  key.foreignport = port2;
  index = insert(&key,ConnHash,&outcome);
  if (index == NOT_FOUND)
    return NOT_FOUND;
  if(outcome == EXISTING)
    return index;

  data = (struct conn_stats*)ConnHash->entries[index].data;

  /* create and initialize conn structure */
  data->sent_to = 0;
  data->received_from = 0;
  data->sent_bytes = 0;
  data->received_bytes = 0;
  data->send_interval = 0;
  data->receive_interval = 0;
  data->classification = TRANSIENT;
  data->old_classification = TRANSIENT;
  data->protocol = proto;
  data->srate = 0;
  data->conn_direction = NONE;
  data->application_state = NOSTATE;     
  data->app_state_expiry = 0;

  ftime(tm);
  data->timestamp = tm->time + (double)tm->millitm/1000;
  data->lst = data->lrt = data->timestamp;
  data->new_connection = TRUE;
  data->ID = CONN_ID ++;
  data->counter = 0;

  if (REPLAY)
    {
      fprintf(conndebug,"%ld: %ld %ld %ld\n",timenow,++conns, false_bad, ConnHash->filled);
      fflush(conndebug);
    }
  else
  if (DEBUG > 1)
    {
      unsigned long curtime = time(0);
      struct in_addr sin, din;
      char source[17], dest[17];
      sin.s_addr = address1;
      din.s_addr = address2;
      strcpy(source,inet_ntoa(sin));
      strcpy(dest,inet_ntoa(din));

      if (index != NOT_FOUND)
	fprintf(conndebug,
		"%ld: Inserted connection %s:%d to %s:%d at index %ld\n",
	         curtime,source,port1,dest,port2,index);
      else
	fprintf(conndebug,
		"%ld: Insert failed for connection %s:%d to %s:%d\n",
		curtime,source,port1,dest,port2);
      fflush(conndebug);
    }
  free(tm);
  connmem -= sizeof(struct timeb);

  update_connections(key.foreignIP,proto);

  /* If hash is close to full and can not grow anymore
     apply drastic cleaning measures to reduce number 
     of full entries */
  if (ConnHash->FULL_ALERT)
    {
      /* We will preferrentialy delete entries that had low
	 number of sent bytes/packets. In first two passes we will 
	 try to avoid as much as we can deletion of good connections */
      int SB = 20;
      int SP = 1;
      int pass = 1;
      if (DEBUG > 1)
	{
	  fprintf(conndebug,
		  "%ld: Cache is %f percent full, raise FULL_ALERT\n", 
		  timenow, (float)ConnHash->filled/ConnHash->size);
	  fflush(conndebug);
	}
      
      /* Do this until occupancy is reduced below high
	 rehash load factor */
      while((double)ConnHash->filled/ConnHash->size 
	    > ConnHash->rehash_lf_high)
	{
	  /* For every entry in hash */
	  for (i=0;i<ConnHash->size;i++)
	    {	
	      /* This is to make sure that we exit as soon as we have enough of free
		 space because otherwise we will be deleting too much */
	      if((double)ConnHash->filled/ConnHash->size 
		 < ConnHash->rehash_lf_high)
		break;
	      /* If there is some data */
	      if (ConnHash->entries[i].state == VALID && index != i)
		{
		  struct conn_key* key = ((struct conn_key*)ConnHash->entries[i].key);
		  struct conn_stats* cs 
		    = (struct conn_stats*)ConnHash->entries[i].data;
		  if (cs->sent_to <= SP 
		      && cs->sent_bytes <=SB 
		      && (cs->classification != GOOD || pass > 2))
		    {
		      if (DEBUG > 1)
			{
			  struct in_addr in;
			  char a1[17],a2[17];
			  in.s_addr = key->localIP;
			  strcpy(a1,inet_ntoa(in));
			  in.s_addr = key->foreignIP;
			  strcpy(a2,inet_ntoa(in));
			  fprintf(conndebug,
				  "%ld: Force deleting entry %ld for connection %s:%d to %s:%d sent %d rec %d class %d\n", 
				  timenow, i, a1,key->localport,
				  a2,key->foreignport,cs->sent_to, cs->received_from,cs->classification);
			  fflush(conndebug);
			}
		      /* We must also delete this flow from good connections */
		      if(cs->classification == GOOD)
			{
			  delconn(key->localIP,key->foreignIP,
				  key->localport, key->foreignport);
			}
		      /* Delete connection from flow statistics */
		      remove_connections(key->foreignIP,cs->protocol);
		      delete(key,ConnHash);
		    }
		}
	    }
	  SB += BINCR;
	  SP++;
	  pass++;
	}
      ConnHash->FULL_ALERT = FALSE;
      /* Now since it is likely that we shuffled everything around return 
	 new index as result of find function */
      index = find(&key,ConnHash);
    }
  return index;
}

/* Get certain fields from connection statistics */
int get_conn(char* address1, char* address2,
		 short port1, short port2, int type)
{
  struct conn_key key;
  struct in_addr in;
  long index;

  gettings++;

  inet_aton(address1,&in);
  key.localIP = in.s_addr;
  inet_aton(address2,&in);
  key.foreignIP = in.s_addr;
  key.localport = port1;
  key.foreignport = port2;

  index = find(&key,ConnHash);

  if (index != NOT_FOUND)  
    {
      struct conn_stats* cs = (struct conn_stats*)ConnHash->entries[index].data;
      switch(type)
	{
	case SENT_PACKETS:
	  return cs->sent_to;
	case REC_PACKETS:
	  return cs->received_from;
	default:
	  return NOT_FOUND;
	}
    }
  return NOT_FOUND;
}

/* When monitoring function captures packet 
   it invokes update_conn to update conn statistics.
   Notice that real statistics for sent packets are
   updated here but will be corrected when rate-limit
   outcome becomes known. Classification module will query 
   this prior to classifying the conn. 
          direction (IN - incoming wrto police address set,
	  OUT - outoging wrto police address set), number of
	  bytes in packet (including packet header),
	  connection protocol */
int update_conn(char* address1, char* address2,
		 int port1, int port2, 
		enum direction dir, int bytes, enum conn_type proto, enum pack_type ptype)
{
  struct in_addr in;  
  unsigned long a1, a2;
  inet_aton(address1,&in);
  a1 = in.s_addr;
  inet_aton(address2,&in);
  a2 = in.s_addr;
  return update_conn_num(a1, a2, port1, port2, dir, bytes, proto, ptype);
}

/* Updates connection statistics for given address
   input: unsigned int IP addresses, short ports
          direction (IN - incoming wrto police address set,
	  OUT - outoging wrto police address set), number of
	  bytes in packet (including packet header), 
	  connection protocol */
int update_conn_num(unsigned int address1,  unsigned int address2, 
		      short port1, short port2, enum direction dir, 
		      int bytes, enum conn_type proto, enum pack_type ptype)
{
  long index;
  struct conn_stats* cs;
  struct timeb* tm = (struct timeb*)malloc(sizeof(struct timeb));
  double curtime;
  struct conn_key key;
  int outcome = 0;
  updates++; 
  connmem += sizeof(struct timeb);

  /* Create connection key */
  key.localIP = address1;
  key.foreignIP = address2;
  key.localport = port1;
  key.foreignport = port2;

  index = insert(&key,ConnHash,&outcome);
  /* If hash is full and we can add no more, give up */
  if (index == NOT_FOUND)
    {
      free(tm);
      connmem -= sizeof(struct timeb);
      return NOT_FOUND;
    }
  
  /* If entry does not exist, add it */
  if (outcome == NEW)
    {
      struct conn_stats* data;

      /* Create and initialize conn structure */
      data = (struct conn_stats*) ConnHash->entries[index].data;
      data->sent_to = 0;
      data->received_from = 0;
      data->sent_bytes = 0;
      data->received_bytes = 0;
      data->send_interval = 0;
      data->receive_interval = 0;
      data->classification = TRANSIENT;
      data->old_classification = TRANSIENT;
      data->protocol = proto;
      data->srate = 0;
      ftime(tm);
      data->timestamp = tm->time + (double)tm->millitm/1000;
      data->lst = data->lrt = data->timestamp;
      data->new_connection = TRUE;
      data->ID = CONN_ID ++; 
      data->conn_direction = dir;
      data->application_state = NOSTATE;     
      if(ptype == DNS || ptype == DNSACK)
	data->app_state_expiry = DNS_EXPIRY;
      else if(ptype == NTP)
	data->app_state_expiry = NTP_EXPIRY;
      else if(ptype == STREAM)
	data->app_state_expiry = 0;
      else
	data->app_state_expiry = 0;
      data->counter = 0;
      inserts++;

      update_connections(key.foreignIP,proto);
    
      /* If hash is close to full and can not grow anymore
	 apply drastic cleaning measures to reduce number 
	 of full entries */
      if (ConnHash->FULL_ALERT)
	{
	  /* We will preferrentialy delete entries that had low
	     number of sent bytes/packets. In first two passes we will 
	     try to avoid as much as we can deletion of good connections */
	  int SB = 20;
	  int SP = 1;
	  int pass = 1;

	  if (DEBUG > 1)
	    {
	      fprintf(conndebug,
		      "%ld: Cache is %f percent full, raise FULL_ALERT\n", 
		      timenow, (float)ConnHash->filled/ConnHash->size);
	      fflush(conndebug);
	    }
	  
	  /* Do this until occupancy is reduced below high
	     rehash load factor */
	  while((double)ConnHash->filled/ConnHash->size 
		> ConnHash->rehash_lf_high)
	    {
	      int i;
	      /* For every entry in hash */
	      for (i=0;i<ConnHash->size;i++)
		{	
		  /* This is to make sure that we exit as soon as we have enough of free
		     space because otherwise we will be deleting too much */
		  if((double)ConnHash->filled/ConnHash->size 
		     < ConnHash->rehash_lf_high)
		    break;
		  /* If there is some data */
		  if (ConnHash->entries[i].state == VALID && index != i)
		    {
		      struct conn_key* key = ((struct conn_key*)ConnHash->entries[i].key);
		      struct conn_stats* cs 
			= (struct conn_stats*)ConnHash->entries[i].data;
		      if (cs->sent_to <= SP 
			  && cs->sent_bytes <=SB 
			  && (cs->classification != GOOD || pass > 2))
			{
			  if (DEBUG > 1)
			    {
			      struct in_addr in;
			      char a1[17],a2[17];
			      in.s_addr = key->localIP;
			      strcpy(a1,inet_ntoa(in));
			      in.s_addr = key->foreignIP;
			      strcpy(a2,inet_ntoa(in));
			      fprintf(conndebug,
				      "%ld: Force deleting entry %d for connection %s:%d to %s:%d sent %d rec %d class %d\n", 
				      timenow, i, a1,key->localport,
				      a2,key->foreignport,cs->sent_to, cs->received_from,cs->classification);
			      fflush(conndebug);
			    }
			  /* We must also delete this flow from good connections */
			  if(cs->classification == GOOD)
			    {
			      delconn(key->localIP,key->foreignIP,
				      key->localport, key->foreignport);
			    }
			  /* Delete connection from flow statistics */
			  remove_connections(key->foreignIP,cs->protocol);
			  delete(key,ConnHash);
			}
		    }
		}
	      SB += BINCR;
	      SP++;
	      pass++;
	    }
	  ConnHash->FULL_ALERT = FALSE;
	  /* Now since it is likely that we shuffled everything around return 
	     new index as result of find function */
	  index = find(&key,ConnHash);
	}
      
      if (REPLAY)
	{
	  fprintf(conndebug,"%ld: %ld %ld %ld\n",timenow,++conns, false_bad, ConnHash->filled);
	  fflush(conndebug);
	}
      else
	if (DEBUG > 1)
	  {
	    unsigned long curtime = time(0);
	    struct in_addr sin, din;
	    char source[17], dest[17];
	    sin.s_addr = address1;
	    din.s_addr = address2;
	    strcpy(source,inet_ntoa(sin));
	    strcpy(dest,inet_ntoa(din));
	  
	    if (index != NOT_FOUND)
	      fprintf(conndebug,
		      "%ld: Inserted connection %s:%d to %s:%d at index %ld\n",
		      curtime,source,port1,dest,port2,index);
	    else
	      fprintf(conndebug,
		      "%ld: Insert failed for connection %s:%d to %s:%d\n",
		    curtime,source,port1,dest,port2);
	    fflush(conndebug);
	  }
    }
  
  /* Update entry */
  cs = (struct conn_stats*)ConnHash->entries[index].data;
  ftime(tm);
  curtime = tm->time + (double)tm->millitm/1000;
  if (dir == IN)
    {
      /* Incoming packet to police address set,
	 update received packet statistics */
      /* Do not count resets */
      if(ptype != RST || cs->protocol != TCP)
	{
	  cs->received_bytes += bytes;
	  if (cs->received_from > 0)
	    cs->receive_interval 
	      = (double)((cs->received_from-1) * cs->receive_interval 
			 + (curtime - cs->lrt))/cs->received_from;
	  else
	    cs->receive_interval = curtime - cs->lrt;
	  cs->lrt = curtime;
	  cs->received_from ++;
	}
      /* Update appliation level models */
      if (cs->protocol == UDP)
	{
	  if (ptype == DNS)
            {
              if(cs->application_state == NOSTATE || cs->application_state == DNS_RESP)
                cs->application_state = DNS_REQ;
              else if((cs->application_state == DNS_REQ) 
		      && (cs->send_interval>cs->app_state_expiry))
                cs->application_state = DNS_REQ;
              else
                cs->application_state = ERROR;
            }
          else if (ptype == DNSACK)
            {
              if (cs->application_state == DNS_REQ)
                cs->application_state = DNS_RESP;
              else
                cs->application_state = ERROR;
            }
	  else if (ptype == NTP)
            {
              if(cs->application_state == NOSTATE)
                cs->application_state = NTP_REC;
              else if(cs->application_state == NTP_SENT)
                cs->application_state = NTP_SENTREC;
              else if((cs->application_state == NTP_SENTREC) &&
		      (cs->receive_interval>cs->app_state_expiry))
                cs->application_state = NTP_SENTREC;
	      else if((cs->application_state == NTP_REC) &&
		      (cs->receive_interval>cs->app_state_expiry))
                cs->application_state = NTP_REC;
              else
                cs->application_state = ERROR;
            }
	  else if (ptype == STREAM)
	    {
	      /* Check if there is RTSP connection that is live */
	      struct conn_stats *rtsp1 = 
		(struct conn_stats*)
		find_partial_conn_num(address1,RTSP_PORT,address2);
	      if (rtsp1)
		{
		  float stale;
		  ftime(tm);
		  /* Check liveness */
		  stale = tm->time + (double)tm->millitm/1000 - cs->timestamp;
		  printf("Found RTSP, stale %f\n",stale);
		  if(stale > STREAM_STALE)
		    cs->application_state = ERROR;
		  else
		    cs->application_state = STREAMING;
		}
	      else
		{
		  cs->application_state = ERROR;
		  printf("Not Found RTSP\n");
		}
	    }
	}
    }
  else
    {	
      /* Sent packet from police address set,
	 update sent packet statistics */
      cs->sent_bytes += bytes;
      if (cs->sent_to > 0)
	cs->send_interval 
	  = (double)((cs->sent_to-1) * cs->send_interval 
		     + (curtime - cs->lst))/cs->sent_to;
      else
	cs->send_interval = curtime - cs->lst;
      cs->lst = curtime;
      cs->sent_to++;
      /* Handle application level stats */
      if (cs->protocol == UDP)
	{
	  if (ptype == DNS)
	    {
	      if(cs->application_state == NOSTATE || cs->application_state == DNS_RESP)
		cs->application_state = DNS_REQ;
              else if((cs->application_state == DNS_REQ) 
		      && (cs->send_interval>cs->app_state_expiry))
		cs->application_state = DNS_REQ;
              else
                cs->application_state = ERROR;
	    }
        else if (ptype == DNSACK)
            {
              if (cs->application_state == DNS_REQ)
                cs->application_state = DNS_RESP;
              else
                cs->application_state = ERROR;
            }
          else if (ptype == NTP)
            {
              if(cs->application_state == NOSTATE)
                cs->application_state = NTP_SENT;
	      else if(cs->application_state == NTP_REC)
                cs->application_state = NTP_SENTREC;
              else if((cs->application_state == NTP_SENTREC) &&
		      (cs->send_interval>cs->app_state_expiry))
                cs->application_state = NTP_SENTREC;
              else if((cs->application_state == NTP_SENT) &&
		      (cs->send_interval>cs->app_state_expiry))
                cs->application_state = NTP_SENT;
              else
                cs->application_state = ERROR;
            }
	  else if (ptype == STREAM)
	    {
	      /* Check if there is RTSP connection that is live */
	      struct conn_stats *rtsp1 = 
		(struct conn_stats*)
		find_partial_conn_num(address1,RTSP_PORT,address2);
	      if (rtsp1)
		{
		  float stale;
		  ftime(tm);
		  /* Check liveness */
		  stale = tm->time + (double)tm->millitm/1000 - cs->timestamp;
		  printf("Found RTSP, stale %f\n",stale);
		  if(stale > STREAM_STALE)
		    cs->application_state = ERROR;
		  else
		    cs->application_state = STREAMING;
		}
	      else
		{
		  cs->application_state = ERROR;
		  printf("Not Found RTSP\n");
		}
	    }
	}
    }
  free(tm);
  connmem -= sizeof(struct timeb);
  return cs->classification;
}

/* Function that deletes hash related variables in case of
   interrupted operaton */
void connhash_cleanup(int signo)
{
  free(ConnHash->entries);
  free(ConnHash);
}

/* Reset statistics for all connections for given 
   destination address; this does not obtain lock since
   calling function does it instead */
void reset_connection(long index)
{
  struct timeb* tm;
  long curtime = time(0);

  struct conn_stats* cs = ((struct conn_stats*)ConnHash->entries[index].data);
  struct conn_key* key = ((struct conn_key*)ConnHash->entries[index].key);

  /* Check if entry is inactive and can be deleted */
  if (conn_clean(cs))
    {
      if (DEBUG > 1)
	{
	  struct in_addr in;
	  char a1[17],a2[17];
	  in.s_addr = key->localIP;
	  strcpy(a1,inet_ntoa(in));
	  in.s_addr = key->foreignIP;
	  strcpy(a2,inet_ntoa(in));
	  fprintf(conndebug,
		  "%ld: Deleting inactive entry for connection %s:%d to %s:%d\n", 
		      curtime, a1,key->localport, a2,key->foreignport);
	  fflush(conndebug);
	}
      /* We must also delete this flow from good connections */
      if(cs->classification == GOOD)
	{
	  delconn(key->localIP,key->foreignIP,
		  key->localport,key->foreignport);
	}
      /* Remove connection from flow statistics */
      remove_connections(key->foreignIP,cs->protocol);
      delete(key,ConnHash);
    }
  else
    {    
      tm = (struct timeb*)malloc(sizeof(struct timeb));  
      connmem += sizeof(struct timeb);
      
      ftime(tm);
      if (DEBUG > 1)
	{
	  struct in_addr in;
	  char a1[17], a2[17];
	  in.s_addr = key->localIP;
	  strcpy(a1,inet_ntoa(in));
	  in.s_addr = key->foreignIP;
	  strcpy(a2,inet_ntoa(in));
	  fprintf(conndebug,
		  "%ld: Reset entry %ld for connection %s:%d %s:%d sent %d rec %d class %d\n", 
		  curtime, index,a1,key->localport,a2,key->foreignport, cs->sent_to,cs->received_from,cs->classification);
	  fflush(conndebug);
	}
      /* Update sending rate in Bytes per second */
      cs->srate = (double)cs->sent_bytes/
	(tm->time + (double)tm->millitm/1000 - cs->timestamp);
      
      /* Reset statistics */
      cs->sent_to = 0;
      cs->received_from = 0;
      cs->sent_bytes = 0;
      cs->received_bytes = 0;
      cs->send_interval = 0;
      cs->receive_interval = 0;
      cs->rto = 0;
      cs->timestamp =  tm->time + (double)tm->millitm/1000;
      
      /* Erase drop from kernel */
      resetconn(key->localIP,key->foreignIP,key->localport,key->foreignport);
      free(tm);
      connmem -= sizeof(struct timeb);
    }
}

/* Initializing statistics hash */
void initialize_conns()
{
  /* Create hash with same min and max size to avoid rehashing
     since we did not use locks for any operation */
  create_hash(&ConnHash, CONN_HASH_SIZE, CONN_HASH_SIZE,
	      CONN_HASH_SIZE, CONN_REHASH_LOAD_FACTOR_HIGH, 
	      CONN_REHASH_LOAD_FACTOR_LOW,  
	      sizeof(struct conn_key),
	      sizeof(struct conn_stats),
	      (int (*)(void*))conn_clean, 
	      (int (*)(void*,void*))conn_compare, 
	      (unsigned long (*)(void*,unsigned long))conn_hash1,
	      (unsigned long (*)(void*,unsigned long))conn_hash2);
}













