/*
Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.*/

/* System includes */
#include <stdio.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/timex.h>
#include <sys/timeb.h>
#include <signal.h>
#include <stdlib.h>
#include <math.h>

/* Our includes */
#include "hash.h"
#include "constants.h"
#include "flow_stats.h"
#include "rl_stats.h"
#include "dward.config"
#include "conn_stats.h"
#include "dward.h"

/* Statistical data */
long flows = 0;
long false_positives = 0;
extern long classifications;

/* Current time */
extern long timenow;

/* Gather data on allocated memory, useful for checking if 
   everything is freed or there is memory leak */
int flowmem = 0;

/* Flow hash */
struct hash_table* FlowHash; 

/* Debug file for flow statistics */
FILE* flowdebug;

/* Debug flag */
extern int DEBUG;

/* Statistics flag */
extern int STATS;

/* GUI flag */
extern int GUI;

/* Flow ID for red-teaming */
int FLOW_ID = 1;

/* Hash related functions we need to define */
int flow_clean(struct flow_stats* data)
{
  /* Hash entry should be cleaned if there is 
     no flow activity for a certain period and 
     flow was characterized as NORMAL */
  long curtime = time(0);
  if (curtime-data->lst > FLOW_INACTIVE_PERIOD && 
      curtime-data->lrt > FLOW_INACTIVE_PERIOD && 
      data->classification == NORMAL)
    return 1;
  else
    return 0;
}

int flow_compare(unsigned int* k1, unsigned int* k2)
{
  if (k1 == NULL || k2 == NULL)
    return FALSE;
  if (*k1==*k2)
    return 1;
  else
    return 0;
}

unsigned long flow_hash1(unsigned int* k, unsigned long size)
{
  return ((*k) % size);
}

unsigned long flow_hash2(unsigned int* k, unsigned long size)
{
  return (1+(*k)%(size-1));
}
/* End of required hash related functions */

/* Both find functions are unsafe to use since they 
   return index to the entry that can be changed
   any moment due to rehashing or parallel access
   by other process */

/* Finds  given IP address in hash, 
   really just masks the fact that 
   we use hash for statistics bookkeeping 
   input: IP address in char array (dotted notation)			       
   output: index (-1 not found)	*/
long find_flow(char *address)
{
  struct in_addr in;  
  unsigned long num;
  inet_aton(address,&in);
  num = in.s_addr;
  return find_flow_num(num); 
}

/* Finds given IP address in numeric form, 
   really just masks the fact that 
   we use hash for statistics bookkeeping 
   input: unsigned int IP address 			       
   output: index (-1 not found)	*/
long find_flow_num(unsigned int address)
{
  long index;
  index = find(&address,FlowHash);
  return index;
}

/* Find flow classification,
   input: IP address in numeric form
   output: flow class, NOT_FOUND for non-existant flow */
int find_flow_class(unsigned int address)
{
  int index;
  enum flow_class classification;
  index = find_flow_num(address);
  if (index != NOT_FOUND)
    classification = 
      ((struct flow_stats*)FlowHash->entries[index].data) ->classification;
  else
    classification = NOT_FOUND;
  return classification;
}

/* Add IP address to hash, return index 
   input: IP address in char array (dotted notation)			       
   output: index */ 
long add_flow(char* address)
{
  struct in_addr in;
  unsigned long num;

  inet_aton(address,&in);
  num = in.s_addr;
  return add_flow_num(num);
}

/* Add IP address to hash, return index 
   input: unsigned int IP address			       
   output: index */ 
long add_flow_num(unsigned int address)
{
  long index, i;
  struct timeb* tm = (struct timeb*)malloc(sizeof(struct timeb));
  struct flow_stats* data;
  int outcome = 0;
  index = insert(&address,FlowHash,&outcome);
  if (index == NOT_FOUND)
    return NOT_FOUND;

  /* Check if flow already exists */
  if (outcome == EXISTING)
    return index;
  data = (struct flow_stats*) FlowHash->entries[index].data;

  /* Create and initialize flow structure */
  data->sent_to = 0;
  data->good_bytes = 0;
  data->received_from = 0;
  data->fw_packets = 0;
  data->fw_bytes = 0;
  data->sent_bytes = 0;
  data->received_bytes = 0;
  data->send_interval = 0;
  data->receive_interval = 0;
  data->classification = NORMAL;
  data->old_classification = NORMAL;
  data->limited = FALSE;
  data->compliant = 0;
  data->TCP_received_from = 0;
  data->TCP_sent_to = 0;
  data->TCP_sent_bytes = 0;
  data->TCP_received_bytes = 0;
  data->UDP_received_from = 0;
  data->UDP_sent_to = 0;
  data->UDP_sent_bytes = 0;
  data->UDP_received_bytes = 0;
  data->ICMP_received_from = 0;
  data->ICMP_sent_to = 0;
  data->ICMP_sent_bytes = 0;
  data->ICMP_received_bytes = 0;
  data->OTHER_received_from = 0;
  data->OTHER_sent_to = 0;
  data->OTHER_sent_bytes = 0;
  data->OTHER_received_bytes = 0;
  data->tcprto = data->icmprto = 0;
  data->tcp = data->udp = data->icmp = data->other = 0;
  data->TCP_conns = data->UDP_conns = data->ICMP_conns = data->OTHER_conns = 0;
  ftime(tm);
  data->timestamp = tm->time + (double)tm->millitm/1000;
  data->lst = data->lrt = data->timestamp;
  data->min_TCP = data->min_UDP = data->min_ICMP = data->min_OTHER = 0;
  data->track_TCP = data->track_UDP = data->track_ICMP = data->track_OTHER = -1;
  data->avg_time = 0;
  data->pos = 0;
  data->DefCOM_attack = FALSE;
  data->DefCOM_rate = -1;
  data->COSSACK_attack = NO_ATTACK;
  data->COSSACK_rate = NOT_FOUND;
  data->COSSACK_type = NO_ATTACK;
  data->COSSACK_new = FALSE;
  data->COSSACK_timestamp = 0;
  data->ID = 0;
  data->observe = 0;

  if (DEBUG)
    {
      unsigned long curtime = time(0);
      struct in_addr in;
      in.s_addr = address;
      
      if (index != NOT_FOUND)
	fprintf(flowdebug,
		"%ld: Inserted address %s at index %ld\n",
		curtime,inet_ntoa(in),index);
      else
	fprintf(flowdebug,
		"%ld: Insert failed for address %s\n",
		curtime,inet_ntoa(in));
      fflush(flowdebug);
    }
  free(tm);
  flowmem -= sizeof(struct timeb);
  
  /* If hash is close to full and can not grow anymore
     apply drastic cleaning measures to reduce number 
     of full entries */
  if (FlowHash->FULL_ALERT)
    {
      /* We will preferrentialy delete entries that had low
	 number of sent bytes/packets */
      int SB = 20;
      int SP = 1;
      
      if (DEBUG > 1)
	{
	  fprintf(flowdebug,
		  "%ld: Cache is %f percent full, raise FULL_ALERT\n", 
		  timenow, (float)FlowHash->filled/FlowHash->size);
	  fflush(flowdebug);
	}

      /* Do this until occupancy is reduced below high
	 rehash load factor */
      while((double)FlowHash->filled/FlowHash->size 
	    > FlowHash->rehash_lf_high)
	{
	  /* For every entry in hash */
	  for (i=0;i<FlowHash->size;i++)
	    {	
	      /* This is so that we don't clean too much */
	      if ((double)FlowHash->filled/FlowHash->size 
		  <= FlowHash->rehash_lf_high)
		break;
	      /* If there is some data */
	      if (FlowHash->entries[i].state == VALID && i != index)
		{
		  struct flow_stats* fs 
		    = (struct flow_stats*)FlowHash->entries[i].data;
		  if (fs->sent_to <= SP 
		      && fs->sent_bytes <=SB)
		    {
		      if (DEBUG > 1)
			{
			  struct in_addr in;
			  in.s_addr = 
			    *((unsigned long*)FlowHash->entries[i].key);
			  fprintf(flowdebug,
				  "%ld: Force deleting entry %ld for address %s\n", 
				  timenow, i, inet_ntoa(in));
			  fflush(flowdebug);
			}
		      /* We must also delete this flow 
			 from rate limit hash */
		      delete_rate_limit_num(*(unsigned int*)
					    FlowHash->entries[i].key);
		      
		      delete(FlowHash->entries[i].key,FlowHash);
		    }
		}
	    }
	  SB += BINCR;
	  SP++;
	}
      /* Remove FULL_ALERT */
      FlowHash->FULL_ALERT = FALSE;
    }
  return index;
}

/* When monitoring function captures packet 
   it invokes update_flow to update flow statistics.
   Notice that real statistics for sent packets are
   updated here but will be corrected when rate-limit
   outcome becomes known. Classification module will query 
   this prior to classifying the flow. 
   input: IP address in char array (dotted notation),
          direction (IN - incoming wrto police address set,
	  OUT - outoging wrto police address set), number of
	  bytes in packet (including packet header) */
void update_flow(char* address, enum direction dir, int bytes, int protocol, int tos, int class, enum pack_type ptype)
{
  struct in_addr in;  
  unsigned long num;
  inet_aton(address,&in);
  num = in.s_addr;
  update_flow_num(num, dir, bytes, protocol, tos, class, ptype);
}

/* Updates statistics for given address
   input: unsigned int IP address
          direction (IN - incoming wrto police address set,
	  OUT - outoging wrto police address set), number of
	  bytes in packet (including packet header) */
void update_flow_num(unsigned int address, enum direction dir, int bytes, int protocol, int tos, int class, enum pack_type ptype)
{
  long index;
  struct flow_stats* fs;
  struct timeb* tm = (struct timeb*)malloc(sizeof(struct timeb));
  double curtime;
  int outcome = 0;
  flowmem += sizeof(struct timeb);

  /* Call insert instead of find and then insert */
  index = insert(&address,FlowHash, &outcome);

  /* If hash is full and we can add no more, give up */
  if (index == NOT_FOUND)
  {
    free(tm);
    flowmem -= sizeof(struct timeb);
    return;
  }

  /* Check if flow already exists */
  if (outcome == NEW)
    {
      struct flow_stats* data;
      int i;
      
      data = (struct flow_stats*) FlowHash->entries[index].data;
      
      /* Create and initialize flow structure */
      data->sent_to = 0;
      data->good_bytes = 0;
      data->fw_packets = 0;
      data->fw_bytes = 0;
      data->received_from = 0;
      data->sent_bytes = 0;
      data->received_bytes = 0;
      data->send_interval = 0;
      data->receive_interval = 0;
      data->classification = NORMAL;
      data->old_classification = NORMAL;    
      data->limited = FALSE;      
      data->compliant = 0;
      data->attack_packets = 0;
      data->attack_bytes = 0;
      data->tcprto = data->icmprto = 0;
      data->TCP_received_from = 0;
      data->TCP_sent_to = 0;
      data->TCP_sent_bytes = 0;
      data->TCP_received_bytes = 0;
      data->UDP_received_from = 0;
      data->UDP_sent_to = 0;
      data->UDP_sent_bytes = 0;
      data->UDP_received_bytes = 0;
      data->ICMP_received_from = 0;
      data->ICMP_sent_to = 0;
      data->ICMP_sent_bytes = 0;
      data->ICMP_received_bytes = 0;
      data->OTHER_received_from = 0;
      data->OTHER_sent_to = 0;
      data->OTHER_sent_bytes = 0;
      data->OTHER_received_bytes = 0;
      data->tcp = data->udp = data->icmp = data->other = 0;
      data->TCP_conns = data->UDP_conns = data->ICMP_conns = data->OTHER_conns = 0;      
      ftime(tm);
      data->timestamp = tm->time + (double)tm->millitm/1000;
      data->lst = data->lrt = data->timestamp;
      data->min_TCP = data->min_UDP = data->min_ICMP = data->min_OTHER = 0;
      data->track_TCP = data->track_UDP = data->track_ICMP = data->track_OTHER = -1;
      data->avg_time = 0;
      data->pos = 0;
      data->DefCOM_attack = FALSE;
      data->DefCOM_rate = -1;
      data->COSSACK_attack = NO_ATTACK;
      data->COSSACK_rate = NOT_FOUND;
      data->COSSACK_type = NO_ATTACK;
      data->COSSACK_new = FALSE;
      data->COSSACK_timestamp = 0;
      data->ID = 0;
      data->observe = 0;

      if (DEBUG)
	{
	  unsigned long curtime = time(0);
	  struct in_addr in;
	  in.s_addr = address;
	    
	  if (index != NOT_FOUND)
	    fprintf(flowdebug,
		    "%ld: Inserted address %s at index %ld\n",
		    curtime,inet_ntoa(in),index);
	  else
	    fprintf(flowdebug,
		    "%ld: Insert failed for address %s\n",
		    curtime,inet_ntoa(in));
	  fflush(flowdebug);
	}
    }
  
  /* If hash is close to full and can not grow anymore
     apply drastic cleaning measures to reduce number 
     of full entries */
  if (FlowHash->FULL_ALERT)
    {
      /* We will preferrentialy delete entries that had low
	 number of sent bytes/packets */
      int SB = 20;
      int SP = 1;
      
      if (DEBUG > 1)
	{
	  fprintf(flowdebug,
		  "%ld: Cache is %f percent full, raise FULL_ALERT\n", 
		  timenow, (float)FlowHash->filled/FlowHash->size);
	  fflush(flowdebug);
	}

      /* Do this until occupancy is reduced below high
	 rehash load factor */
      while((double)FlowHash->filled/FlowHash->size 
	    > FlowHash->rehash_lf_high)
	{
	  int i;
	  /* For every entry in hash */
	  for (i=0;i<FlowHash->size;i++)
	    {	
	      /* This is so that we don't clean too much */
	      if ((double)FlowHash->filled/FlowHash->size 
		  <= FlowHash->rehash_lf_high)
		break;
	      /* If there is some data */
	      if (FlowHash->entries[i].state == VALID && i != index)
		{
		  struct flow_stats* fs 
		    = (struct flow_stats*)FlowHash->entries[i].data;
		  if (fs->sent_to <= SP 
		      && fs->sent_bytes <=SB)
		    {
		      if (DEBUG > 1)
			{
			  struct in_addr in;
			  in.s_addr = 
			    *((unsigned long*)FlowHash->entries[i].key);
			  fprintf(flowdebug,
				  "%ld: Force deleting entry %d for address %s\n", 
				  timenow, i, inet_ntoa(in));
			  fflush(flowdebug);
			}
		      /* We must also delete this flow 
			 from rate limit hash */
		      delete_rate_limit_num(*(unsigned int*)
					    FlowHash->entries[i].key);
		      
		      delete(FlowHash->entries[i].key,FlowHash);
		    }
		}
	    }
	  SB += BINCR;
	  SP++;
	}
      /* Remove FULL_ALERT */
      FlowHash->FULL_ALERT = FALSE;
    }

  /* Update entry */
  fs = FlowHash->entries[index].data;

  ftime(tm);
  curtime = tm->time + (double)tm->millitm/1000;
  if (dir == IN)
    {

      /* Incoming packet to police address set,
	 update received packet statistics */
      fs->received_bytes += bytes;
      if (fs->received_from > 0)
	fs->receive_interval 
	  = (double)((fs->received_from-1) * fs->receive_interval 
		     + (curtime - fs->lrt))/fs->received_from;
      else
	fs->receive_interval = curtime - fs->lrt;
      fs->lrt = curtime;
      fs->received_from ++;
      if (protocol == TCP)
	{
	  fs->TCP_received_bytes += bytes;
	  fs->TCP_received_from ++;
	  if (ptype == RST)
	{
		fs->received_from--;
		fs->TCP_received_from--;
	}
	}
      else if (protocol == UDP)
	{
	  fs->UDP_received_bytes += bytes;
	  fs->UDP_received_from ++;
	}
      else if (protocol == ICMP)
	{
	  fs->ICMP_received_bytes += bytes;
	  fs->ICMP_received_from ++;
	}
      else
	{
	  fs->OTHER_received_bytes += bytes;
	  fs->OTHER_received_from ++;
	}
    }
  else
    {	
      /* Sent packet from police address set,
	 update sent packet statistics */

      if (fs->sent_to > 0)
	fs->send_interval 
	  = (double)((fs->sent_to-1) * fs->send_interval 
		     + (curtime - fs->lst))/fs->sent_to;
      else
	fs->send_interval = curtime - fs->lst;

      if (class == GOOD)
	fs->good_bytes += bytes;
      fs->lst = curtime;
      if (tos == 255)
	{
	  fs->attack_packets++;
	  fs->attack_bytes += bytes;
	}
      /* Increase counters */
      fs->sent_to++;
      fs->sent_bytes += bytes;
      fs->fw_packets++;
      fs->fw_bytes += bytes;
      if (protocol == TCP)
	{
	  fs->TCP_sent_bytes += bytes;
	  fs->TCP_sent_to ++;
	  if (class != GOOD)
	    fs->min_TCP += bytes;
	}
      else if (protocol == UDP)
	{
	  fs->UDP_sent_bytes += bytes;
	  fs->UDP_sent_to ++;
	  if (class != GOOD)
	    fs->min_UDP += bytes;
	}
      else if (protocol == ICMP)
	{
	  fs->ICMP_sent_bytes += bytes;
	  fs->ICMP_sent_to ++;
	  if (class != GOOD)
	    fs->min_ICMP += bytes;
	}
      else
	{
	  fs->OTHER_sent_bytes += bytes;
	  fs->OTHER_sent_to ++;
	  if (class != GOOD)
	    fs->min_OTHER += bytes;
	}
    }
  free(tm);
  flowmem -= sizeof(struct timeb);
}


/* Sets/resets  COSSACK flags for given address
   input: unsigned int IP address
          COSSACK flag (TRUE/FALSE) */
int set_cossack_flag_and_rate(unsigned int address, int flag, 
			       int rate, int type)
{
  long index;
  struct flow_stats* fs;
  index = find(&address,FlowHash);
  /* If entry does not exist, return.
     This should never happen */
  if (index == NOT_FOUND)
    {
      return 0;
    }
  else
    {
      struct flow_stats *fs = (struct flow_stats*) FlowHash->entries[index].data;
      /* If this is not duplicate action,
	 due to large subnet info */
      if (!fs->COSSACK_new)
	{
	  fs->COSSACK_attack = flag;
	  fs->COSSACK_rate = rate;
	  fs->COSSACK_type = type;
	  fs->COSSACK_new = TRUE;
	}
      fs->COSSACK_timestamp = COSSACK_TIMESTAMP;
      return 1;
    }
}

/* Sets attack alert based on message from DefCOM */
int set_alert(unsigned int address)
{
  long index;
  struct flow_stats* fs;
  index = find(&address,FlowHash);
  /* If entry does not exist, return.
     This should never happen */
  if (index == NOT_FOUND)
    {
      return 0;
    }
  else
    {
      struct flow_stats *fs = (struct flow_stats*) FlowHash->entries[index].data;
      /* If this is not duplicate action,
	 due to large subnet info */
      fs->DefCOM_attack = TRUE;
      return 1;
    }
}


/* Sets rate limit based on message from DefCOM */
int set_rate_limit(unsigned int address, int rate)
{
  long index;
  struct flow_stats* fs;
  index = find(&address,FlowHash);
  /* If entry does not exist, return.
     This should never happen */
  if (index == NOT_FOUND)
    {
      return 0;
    }
  else
    {
      struct flow_stats *fs = (struct flow_stats*) FlowHash->entries[index].data;
      /* If this is not duplicate action,
	 due to large subnet info */
      fs->DefCOM_rate = rate;
      return 1;
    }
}


/* Reset statistics for a given flow;
   this does not obtain lock since calling function
   does it instead */
void reset_flow(long index)
{
  long curtime = time(0);
  struct timeb* tm;
  struct flow_stats* fs;
  int i;
  unsigned int address = *((unsigned int*) FlowHash->entries[index].key);
  fs = (struct flow_stats*) FlowHash->entries[index].data;
  /* Check if entry is inactive and can be deleted */
  if (flow_clean(fs))
    {
      if (DEBUG > 1)
	{
	  struct in_addr in;
	  in.s_addr = address;
	  fprintf(flowdebug,
		  "%ld: Deleting inactive entry for address %s\n", 
		  curtime, inet_ntoa(in));
	  fflush(flowdebug);
	}
      /* We must also delete this flow from rate limit hash */
      delete_rate_limit_num(address);
      delete(&address,FlowHash);
    }
  else
    {
      unsigned long sum;
      tm = (struct timeb*)malloc(sizeof(struct timeb));
      flowmem += sizeof(struct timeb);
      
      if (DEBUG > 1)
	{
	  struct in_addr in;
	  in.s_addr = address;
	  fprintf(flowdebug,
		  "%ld: Reset entry for address %s\n", 
		  curtime, inet_ntoa(in));
	  fflush(flowdebug);
	}
      /* Check if we should send statistics to GUI */
      if (GUI)
	send_to_GUI(fs->fw_packets, fs->received_from, 
		    fs->fw_bytes, fs->received_bytes, 
		    fs->attack_packets, fs->attack_bytes, address);
      /* Reset statistics */
      fs->sent_to = 0;
      fs->good_bytes = 0;
      fs->fw_packets = 0;
      fs->fw_bytes = 0;
      fs->received_from = 0;
      fs->sent_bytes = 0;
      fs->received_bytes = 0;
      fs->send_interval = 0;
      fs->receive_interval = 0;
      fs->attack_packets = 0;
      fs->attack_bytes = 0;
      fs->min_TCP = fs->min_UDP = fs->min_ICMP = fs->min_OTHER = 0;
      if (fs->classification != ATTACK)
	{
	  /* Update only if flow is well-behaved */
	  sum = fs->TCP_sent_bytes + fs->UDP_sent_bytes 
	    + fs->ICMP_sent_bytes + fs->OTHER_sent_bytes;
	  if (sum)
	    fs->tcp = fs->tcp * ALPHA + (1-ALPHA) 
	      * (double)fs->TCP_sent_bytes/sum;
	  else
	    fs->tcp = fs->tcp * ALPHA;
	  if (sum)
	    fs->udp = fs->udp * ALPHA + (1-ALPHA) 
	      * (double)fs->UDP_sent_bytes/sum;
	  else
	    fs->udp = fs->udp * ALPHA;
	  if (sum)
	    fs->icmp = fs->icmp * ALPHA + (1-ALPHA) 
	      * (double)fs->ICMP_sent_bytes/sum;
	  else
	    fs->icmp = fs->icmp * ALPHA;
	  if (sum)
	    fs->other = fs->other * ALPHA + (1-ALPHA) 
	      * (double)fs->OTHER_sent_bytes/sum;
	  else
	    fs->other = fs->other * ALPHA;
	}
      fs->TCP_received_from = 0;
      fs->TCP_sent_to = 0;
      fs->TCP_sent_bytes = 0;
      fs->TCP_received_bytes = 0;
      fs->UDP_received_from = 0;
      fs->UDP_sent_to = 0;
      fs->UDP_sent_bytes = 0;
      fs->UDP_received_bytes = 0;
      fs->ICMP_received_from = 0;
      fs->ICMP_sent_to = 0;
      fs->ICMP_sent_bytes = 0;
      fs->ICMP_received_bytes = 0;
      fs->OTHER_received_from = 0;
      fs->OTHER_sent_to = 0;
      fs->OTHER_sent_bytes = 0;
      fs->OTHER_received_bytes = 0;
      ftime(tm);
      fs->timestamp =  tm->time + (double)tm->millitm/1000;
      /* Erase drops from kernel */
      resetflow(address);
      free(tm);
      flowmem -= sizeof(struct timeb);
    }
}

/* Add new connection to the flow */
void update_connections(unsigned int address, int protocol)
{
  int index = find(&address,FlowHash);

  if (index != NOT_FOUND)
    {
      struct flow_stats* fs 
	= (struct flow_stats*)(FlowHash->entries[index].data);
      switch(protocol)
	{
	case TCP:
	  fs->TCP_conns ++;
	  break;
	case UDP:
	  fs->UDP_conns ++;
	  break;
	case ICMP:
	  fs->ICMP_conns ++;
	  break;
	case OTHER:
	  fs->OTHER_conns ++;
	  break;
	default:
	  break;
	}
    }
}

void monitor_flow(unsigned int address)
{
  int index = find(&address,FlowHash);
  if (index != NOT_FOUND)
    {
      struct flow_stats* fs 
	= (struct flow_stats*)(FlowHash->entries[index].data);
      fs->observe = 1;
    }
}

/* Remove connection from the flow */
void remove_connections(unsigned int address, int protocol)
{
  int index = find(&address,FlowHash);
  if (index != NOT_FOUND)
    {
      struct flow_stats* fs 
	= (struct flow_stats*)(FlowHash->entries[index].data);
      switch(protocol)
	{
	case TCP:
	  fs->TCP_conns --;
	  if (fs->TCP_conns < 0)
	    fs->TCP_conns = 0;
	  break;
	case UDP:
	  fs->UDP_conns --;
	  if (fs->UDP_conns < 0)
	    fs->UDP_conns = 0;
	  break;
	case ICMP:
	  fs->ICMP_conns --;
	  if (fs->ICMP_conns < 0)
	    fs->ICMP_conns = 0;
	  break;
	case OTHER:
	  fs->OTHER_conns --;
	  if (fs->OTHER_conns < 0)
	    fs->OTHER_conns = 0;
	  break;
	default:
	  break;
	}
    }
}

/* Function that deletes hash related variables in case of
   interrupted operaton */
void flowhash_cleanup(int signo)
{
  free(FlowHash->entries);
  free(FlowHash);
}

/* Initializing flow statistics hash */
void initialize_flows()
{
  /* Create hash with same min and max size to avoid rehashing
     since we did not use locks for any operation */
  create_hash(&FlowHash, FLOW_HASH_SIZE, FLOW_HASH_SIZE,
	      FLOW_HASH_SIZE, FLOW_REHASH_LOAD_FACTOR_HIGH, 
	      FLOW_REHASH_LOAD_FACTOR_LOW,  
	      sizeof(unsigned int),
	      sizeof(struct flow_stats),
	      (int (*)(void*))flow_clean, 
	      (int (*)(void*,void*))flow_compare, 
	      (unsigned long (*)(void*,unsigned long))flow_hash1,
	      (unsigned long (*)(void*,unsigned long))flow_hash2);
}












