/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* File that contains some constants for statistics gathering
   on connections. Connection is defined as pair of IP addresses 
   and port numbers. */

#ifndef __CONN_STATS_H
#define __CONN_STATS_H

#include "constants.h"

/* Connection classification */
enum conn_class {GOOD, TRANSIENT, BAD};

/* Type of connection protocol */
enum conn_type {TCP=6, UDP=17, ICMP=1, OTHER=0};

/* Application state */
enum app_state {NOSTATE, ERROR, DNS_REQ, DNS_RESP, 
		NTP_SENT, NTP_REC, NTP_SENTREC, STREAMING};

/* Staleness of RTSP connection */
#define STREAM_STALE 10

/* DNS Expiry */
#define DNS_EXPIRY 1

/* NTP Expiry */
#define NTP_EXPIRY 60

/* UDP application ports */
#define DNS_PORT 53
#define NTP_PORT 123
#define RTSP_PORT 554
#define RP_PORT 6970

struct conn_stats
{
  /* Number of packets received from foreign peer */
  int received_from;		
  /* Number of packets sent to foreign peer */
  int sent_to;			
  /* Number of bytes received from foreign peer */
  int received_bytes;		
  /* Number of bytes sent to foreign peer */
  int sent_bytes;		
  /* Interarrival time between successive packets 
     sent to foreign peer */
  double send_interval;		
  /* Interarrival time between successive packets 
     received from foreign peer */ 
  double receive_interval;	
  /* Last send time within interval to foreign peer */   
  double lst;			
  /* Last receive time within interval from foreign peer */
  double lrt;		
  /* Smoothed sent/rec packet ratio */
  double rto;
  /* Sending rate during previous rate limit */
  int srate;
  /* Time when this entry was last reset */
  long timestamp;
  /* Current connection classification */
  enum conn_class classification;
  /* Old connection classification, 
     for statistics gathering */
  enum conn_class old_classification;
  /* Connection protocol */
  enum conn_type protocol;
  /* Dropped packets */
  int droppedp;
  /* Is this a new connection */
  int new_connection;
  /* For red-teaming */
  int ID;
  /* TCP state */
  enum app_state application_state;
  int counter;
  /* Who initiated connection */
  enum direction conn_direction;
  /* When does application state expire */
  double app_state_expiry;
};

/* Here we define type for connection key that 
   will be used for hashing */
struct conn_key
{
  unsigned int localIP;
  unsigned int foreignIP;
  unsigned short localport;
  unsigned short foreignport;
};

/* Function prototypes, refer to conn_stats.c for descriptions */
int conn_clean(struct conn_stats*);
int conn_compare(struct conn_key*, struct conn_key*);
unsigned long conn_hash1(struct conn_key*, unsigned long);
unsigned long conn_hash2(struct conn_key*, unsigned long);
void clean_conn_hash();
long find_conn(char *, char*, char*, char*);
long find_conn_num(unsigned int,  unsigned int, short, short);
struct conn_stats* find_partial_conn_num(unsigned int,  short, unsigned int);
long add_conn(char*, char*, char*, char*, enum conn_type);
long add_conn_num(unsigned int,  unsigned int, short, short, enum conn_type);
int get_conn(char*, char*, short, short, int);
int update_conn(char*, char*, int, int, enum direction, 
		 int, enum conn_type, enum pack_type);
int update_conn_num(unsigned int,  unsigned int, short, short, 
		     enum direction, int, enum conn_type, enum pack_type);
void initialize_conns();
void connhash_cleanup(int);
void reset_connection(long);
char* print_proto(int);

#endif







