/* Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.*/

/* System includes */
#include <linux/config.h> 
#include <linux/slab.h> 
#include <linux/mm.h> 
#include <linux/module.h>
#include <linux/kernel.h> 
#include <asm/types.h> 
#include <linux/stddef.h> 
#include <linux/types.h> 
#include <linux/netfilter.h> 
#include <linux/netfilter_ipv4.h> 
#include <linux/skbuff.h> 
#include <linux/ip.h> 
#include <linux/tcp.h> 
#include <linux/icmp.h> 
#include <linux/netdevice.h> 
#include <linux/if_ether.h> 
#include <sys/syscall.h>  
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/time.h>
#include <linux/types.h> 
#include <linux/netfilter.h> 
#include <linux/netfilter_ipv4.h> 
#include <linux/skbuff.h>

/* Our includes */
#include "khash.h"
#include "constants.h"
#include "fhash.h"
#include "dward.config"

/* Flow hash */
struct hash_table* FlowHash; 

/* Hash related functions we need to define */
int flow_clean(struct flow_stats* data)
{
  /* Never clean unless explicitly demanded */
    return 0;
}

int flow_compare(unsigned int* k1, unsigned int* k2)
{
  if (k1 == NULL || k2 == NULL)
    return 0;
  if (*k1==*k2)
    return 1;
  else
    return 0;
}

unsigned long flow_hash1(unsigned int* k, unsigned long size)
{
  return ((*k) % size);
}

unsigned long flow_hash2(unsigned int* k, unsigned long size)
{
  return (1+(*k)%(size-1));
}

/* Function that creates non-rehashable hash table to avoid lock delays */
void create_flowhash()
{
  /* Create hash with same min and max size to avoid rehashing
     since we did not use locks for any operation */
  create_hash(&FlowHash, KERNEL_FLOW_HASH_SIZE, 
	      KERNEL_FLOW_HASH_SIZE,
	      KERNEL_FLOW_HASH_SIZE, 
	      KERNEL_FLOW_REHASH_LOAD_FACTOR_HIGH, 
	      KERNEL_FLOW_REHASH_LOAD_FACTOR_LOW,  
	      sizeof(unsigned int),
	      sizeof(struct flow_stats),
	      (int (*)(void*))flow_clean, 
	      (int (*)(void*,void*))flow_compare, 
	      (unsigned long (*)(void*,unsigned long))flow_hash1,
	      (unsigned long (*)(void*,unsigned long))flow_hash2);
}

/* Utility functions that mask hash structure */

/* Get various flow statistics. This is a system call that 
   delivers more accurate statistics from  kernel to application level */
int get_flow(unsigned int address, int type)
{
  int index = find(&address,FlowHash);
  if (index == NOT_FOUND)
    return 0;
  else
    {
      struct flow_stats* fs 
	= (struct flow_stats*)(FlowHash->entries[index].data);
      switch(type)
	{
	case RL:
	  return fs->rate_limit;
	case SENT_PACKETS:
	  return fs->sum_sentp;
	case SENT_BYTES:
	  return fs->sum_sentB;
	case REC_PACKETS:
	  return fs->sum_recp;
	case REC_BYTES:
	  return fs->sum_recB;
	case DROP_PACKETS:
	  return fs->sum_droppedp;
	case DROP_BYTES:
	  return fs->sum_droppedB;
	case SENT_BYTES_SEC:
	  return fs->sentB;
	case GOOD_BYTES:
	  return fs->goodB_old;
	case GOOD_BYTES_SEC:
	  return fs->goodB;
	case ATT_PACKETS:
	  return fs->sum_attp;
	case ATT_BYTES:
	  return fs->sum_attB;
	case FW_PACKETS:
	  return fs->sum_fwp;
	case FW_BYTES:
	  return fs->sum_fwB;
	case SYNS:
	  return fs->syns_sent;
	default:
	  return 0;
	}
    }
}

/* Find flow in hash and return its index */
long find_flow(unsigned int address)
{
  long index;
  index = find(&address, FlowHash);
  return index;
}

/* Insert the desired rate limit for the flow. Special value of -1
   is inserted if the application cannot decide on the rate limit but
   wants the kernel to estimate the rate and then insert half of that value
   as rate limit */
long insert_rate_limit(unsigned int address, int rl)
{
  int outcome;
  struct flow_stats* data;
  int index = insert(&address, FlowHash, &outcome);

  if (index!=NOT_FOUND && outcome == NEW)
      {
	struct timeval tv;
	/* This is newly rate-limited flow, the rl argument should be -1 */
	data = (struct flow_stats*)FlowHash->entries[index].data;
	data->rate_limit = rl;
	data->syns_sent = 0;
	data->next_rate_limit = 0;
	data->sum_sentp = 0;
	data->sum_sentB = 0;
	data->sum_recp = 0;
	data->sum_recB = 0;
	data->sum_attp = 0;
	data->sum_attB = 0;
	data->sum_fwp = 0;
	data->sum_fwB = 0;
	data->sum_droppedp = 0;
	data->sum_droppedB = 0;
	data->sentp = 0;
	data->sentB = 0;
	data->recp = 0;
	data->recB = 0;
	data->attp = 0;
	data->attB = 0;
	data->fwp = 0;
	data->fwB = 0;
	data->droppedp = 0;
	data->droppedB = 0;
	data->goodB = 0;
	data->goodB_old = 0;
	do_gettimeofday(&tv);
	data->timestamp_sec = tv.tv_sec;
	data->timestamp_usec = tv.tv_usec;
	return index;
      }
  else if(index!=NOT_FOUND && outcome == EXISTING)
      {
	/* Just apply the desired rate-limit */
	struct flow_stats* fs 
	  = (struct flow_stats*)(FlowHash->entries[index].data);
	fs->next_rate_limit = rl;
	return index;
      }
  return NOT_FOUND;
}

/* Update flow with new packet arrival */
void update_flow(unsigned int address, int bytes, 
		 enum direction dir,int attack, int syn)
{
  int outcome;
  struct flow_stats* fs;
  int index = insert(&address, FlowHash,&outcome);
  if(index != NOT_FOUND && outcome == EXISTING)
    {
      	struct flow_stats* fs 
	  = (struct flow_stats*)(FlowHash->entries[index].data);
	/* Update all temporary values */
	if (dir == OUT)
	  {
	    if (syn)
	      fs->syns_sent += bytes;
	    fs->sentp ++;
	    fs->sentB += bytes;
	    if (attack)
	      {
		fs->attp ++;
		fs->attB += bytes;
	      }
	    fs->fwp ++;
	    fs->fwB += bytes;
	  }
	else
	  {
	    fs->recp ++;
	    fs->recB += bytes;
	  }
    }
}

/* Update flow with new packet drop */
void update_flow_drop(unsigned int address, int bytes, int gflag)
{
  int index = find(&address, FlowHash);
  if(index != NOT_FOUND)
    {
      	struct flow_stats* fs 
	  = (struct flow_stats*)(FlowHash->entries[index].data);
	/* Update temporary drop values */
	fs->droppedp ++;
	fs->droppedB += bytes;
    }
}

/* Update flow with goodput */
void update_flow_good(unsigned int address, int bytes)
{
  int index = find(&address, FlowHash);
  if(index != NOT_FOUND)
    {
      	struct flow_stats* fs 
	  = (struct flow_stats*)(FlowHash->entries[index].data);
	/* Update temporary drop values */
	fs->goodB += bytes;
    }
}

/* Reset flow statistics */
void reset_flow(unsigned int address)
{
  int index = find(&address, FlowHash);
  if(index != NOT_FOUND)
    {
      /* Reset all cumulative measures */
      	struct flow_stats* fs 
	  = (struct flow_stats*)(FlowHash->entries[index].data);
	fs->sum_sentp = 0;
	fs->sum_recp = 0;
	fs->sum_sentB = 0;
	fs->sum_recB = 0;
	fs->sum_attp = 0;
	fs->sum_attB = 0;
	fs->sum_fwp = 0;
	fs->sum_fwB = 0;
	fs->sum_droppedp = 0;
	fs->sum_droppedB = 0;
    }
}

/* Delete flow from the hash */
void delete_rate_limit(unsigned int address)
{
  delete(&address,FlowHash);
}

/* Reset one-second statistics for all flows */
void reset_flows()
{
  int i;
  /* Add all one-second statistics onto cumulative statistics and 
     then reset one-second counters */
  for(i=0;i<FlowHash->size;i++)
    {
      if(FlowHash->entries[i].state == VALID)
	{
	  struct flow_stats* fs 
	  = (struct flow_stats*)(FlowHash->entries[i].data);
	  struct timeval tv;
	  fs->syns_sent = 0;
	  fs->sum_sentp += fs->sentp;
	  fs->sum_sentB += fs->sentB;
	  fs->sum_recp += fs->recp;
	  fs->sum_recB += fs->recB;
	  fs->sum_attp += fs->attp;
	  fs->sum_attB += fs->attB;
	  fs->sum_fwp += fs->fwp;
	  fs->sum_fwB += fs->fwB;
	  fs->sum_droppedp += fs->droppedp;
	  fs->sum_droppedB += fs->droppedB;
	  /* If rate limit was -1 update it to be 0, 
	     in next interval the rate limit will be set 
	     to a half of the measured rate */
	  if (fs->rate_limit == NOT_FOUND)
	    fs->rate_limit = 0;
	  else if (fs->rate_limit == 0 && fs->sentB > 0)
	    fs->rate_limit = fs->sentB/2;
	  /* Shift measure of good connections throughput */
	  if (fs->goodB_old == 0)
	    fs->goodB_old = fs->goodB;
	  else if (fs->goodB >= fs->goodB_old)
	    fs->goodB_old = fs->goodB;
	  else
	    fs->goodB_old *= 0.99;
	  /* Transfer next rate limit */
	  if (fs->next_rate_limit != 0)
	    {
	      fs->rate_limit = fs->next_rate_limit;
	      fs->next_rate_limit = 0;
	    }
	  /* Adjust if not between MIN_RATE and MAX_RATE */
	  if (fs->rate_limit > 0 && fs->rate_limit < MIN_RATE)
	    fs->rate_limit = MIN_RATE;
	  if (fs->rate_limit > MAX_RATE)
	    fs->rate_limit = MAX_RATE;
	  /* Update timestamp */
	  do_gettimeofday(&tv);
	  fs->timestamp_sec = tv.tv_sec;
	  fs->timestamp_usec = tv.tv_usec;
	  /* Reset all one-second statistics */
	  fs->sentp = 0;
	  fs->sentB = 0;
	  fs->recp = 0;
	  fs->recB = 0;
	  fs->attp = 0;
	  fs->attB = 0;
	  fs->fwp = 0;
	  fs->fwB = 0;
	  fs->droppedp = 0;
	  fs->droppedB = 0;
	  fs->goodB = 0;
	}
    }
}

/* Delete hash table on exit */
void delete_flowhash()
{
  int i;
  if (FlowHash == NULL || FlowHash->entries == NULL)
    return;
  for(i=0;i<FlowHash->size;i++)
    {
      if(FlowHash->entries[i].state == VALID)
	{
	  kfree(FlowHash->entries[i].key);	 
	  FlowHash->entries[i].key = NULL;
	  kfree(FlowHash->entries[i].data);
	  FlowHash->entries[i].data = NULL;
	}
    }
  kfree(FlowHash->entries);
  FlowHash->entries = NULL;
  kfree(FlowHash);
  FlowHash = NULL;
}









